# (c) Copyright 2014. CodeWeavers, Inc.

import os.path
import re
import sys

import globtree
import cxlog
import cxutils
import distversion

# for localization
from cxutils import cxgettext as _


# Notes:
# - One day we may want to store these in an automatically updated XML file.
# - Fedora checks that the 32 and 64 bit packages have the same version. So
#   always tell yum to install/update both.
FIXES = \
    {'dbus':
         {'': _('Could not load the D-Bus Python modules.'),
          'packages': {
                ('debian6', None):  ('python-dbus',),
                ('debian7', None):  ('python-dbus',),
                ('fedora', None):   ('dbus-python',),
                ('mandriva', None): ('dbus-python',),
                ('suse', None):     ('dbus-1-python',),
                },
          },
     'gtk2':
         {'': _('Could not load the GTK+ Python modules.'),
          'packages': {
                ('debian6', None):  ('python-gtk2',),
                ('debian7', None):  ('python-gtk2',),
                ('fedora', None):   ('pygtk2',),
                ('mandriva', None): ('pygtk2.0',),
                ('suse', None):     ('python-gtk',),
                },
          },

     # The critical 32 bit libraries needed by Wine

     'missinglibc':
         {'': _('Could not find the 32-bit C library.'),
          'packages': {
                ('debian6', '64'):  ('libc6-i386',),
                ('debian7', '64'):  ('libc6-i386',),
                ('fedora', '64'):   ('glibc.i686', 'glibc.x86_64', 'zlib.i686', 'zlib.x86_64'),
                ('suse', '64'):     ('glibc-32bit', 'zlib-32bit'),
                },
          },
     'missinglibgcc_s':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libgcc1',),
                ('debian6', '64'):  ('lib32gcc1',),
                ('debian7', None):  ('libgcc1:i386',),
                ('fedora', '32'):   ('libgcc1',),
                ('fedora', '64'):   ('libgcc.i686', 'libgcc.x86_64'),
                ('mandriva', None): ('libgcc1',),
                ('suse', '32'):     ('libgcc1',), # Virtual package
                ('suse', '64'):     ('libgcc1-32bit',), # Virtual package
                },
          },
     'missinglibnss_mdns':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('centos6', None):  tuple(),
                ('debian6', '32'):  ('libnss-mdns',),
                ('debian6', '64'):  ('lib32nss-mdns',),
                ('debian7', '32'):  ('libnss-mdns',),
                ('debian7', '64'):  ('lib32nss-mdns',),
                ('fedora', '32'):   ('nss-mdns',),
                ('fedora', '64'):   ('nss-mdns.i686', 'nss-mdns.x86_64'),
                ('mandriva', None): ('nss_mdns',),
                ('suse', '32'):     ('nss-mdns',),
                ('suse', '64'):     ('nss-mdns-32bit',),
                ('ubuntu', None):   ('libnss-mdns:i386',),
                },
          },
     'missinglibice':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libice6',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libice6:i386',),
                ('fedora', '32'):   ('libICE',),
                ('fedora', '64'):   ('libICE.i686', 'libICE.x86_64'),
                ('mandriva', None): ('libice6',),
                ('suse', '32'):     ('libICE6',),
                ('suse', '64'):     ('libICE6-32bit',),
                },
          },
     'missinglibsm':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libsm6',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libsm6:i386',),
                ('fedora', '32'):   ('libSM',),
                ('fedora', '64'):   ('libSM.i686', 'libSM.x86_64'),
                ('mandriva', None): ('libsm6',),
                ('suse11', '32'):   ('xorg-x11-libICE',),
                ('suse11', '64'):   ('xorg-x11-libICE-32bit',),
                ('suse', '32'):     ('libSM6',),
                ('suse', '64'):     ('libSM6-32bit',),
                },
          },
     'missinglibx11':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libx11-6',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libx11-6:i386',),
                ('fedora', '32'):   ('libX11',),
                ('fedora', '64'):   ('libX11.i686', 'libX11.x86_64'),
                ('mandriva', None): ('libx11_6',),
                ('suse11', '32'):   ('xorg-x11-libX11',),
                ('suse11', '64'):   ('xorg-x11-libX11-32bit',),
                ('suse', '32'):     ('libX11-6',),
                ('suse', '64'):     ('libX11-6-32bit',),
                },
          },
     'missinglibxext':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libxext6',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxext6:i386',),
                ('fedora', '32'):   ('libXext',),
                ('fedora', '64'):   ('libXext.i686', 'libXext.x86_64'),
                ('mandriva', None): ('libxext6',),
                ('suse11', '32'):   ('xorg-x11-libXext',),
                ('suse11', '64'):   ('xorg-x11-libXext-32bit',),
                ('suse', '32'):     ('libXext6',),
                ('suse', '64'):     ('libXext6-32bit',),
                },
          },
     'missinglibfreetype':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libfreetype6',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libfreetype6:i386',),
                ('fedora', '32'):   ('freetype',),
                ('fedora', '64'):   ('freetype.i686', 'freetype.x86_64'),
                ('mandriva', None): ('libfreetype6',),
                ('suse', '32'):     ('freetype2',), # Virtual package
                ('suse', '64'):     ('freetype2-32bit',), # Virtual package
                },
          },
     'missinglibpng':
         {'tiedto': set(('missinglibc',)),
          'packages': {
                ('debian6', '32'):  ('libpng12-0',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libpng12-0:i386',),
                ('fedora', '32'):   ('libpng',),
                ('fedora', '64'):   ('libpng.i686', 'libpng.x86_64'),
                ('mandriva', None): ('libpng3',),
                ('suse', '32'):     ('libpng',),
                ('suse', '64'):     ('libpng12-0-32bit',),
                },
          },

     # Other 32 bit libraries used by Wine

     'missinglibasound':
         {'packages': {
                ('debian6', '32'):  ('libasound2',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libasound2:i386',),
                ('fedora', '32'):   ('alsa-lib',),
                ('fedora', '64'):   ('alsa-lib.i686', 'alsa-lib.x86_64'),
                ('mandriva', None): ('libalsa2',),
                ('suse', '32'):     ('libasound2',),
                ('suse', '64'):     ('libasound2-32bit',),
                },
          },
     'missinglibcapi20':
         {'packages': {
                ('debian6', '32'):  ('libcapi20-3',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libcapi20-3:i386',),
                ('fedora', '32'):   ('isdn4k-utils',),
                ('fedora', '64'):   ('isdn4k-utils.i686',),
                ('mandriva', None): ('libisdn4k-utils3',),
                ('suse11', '32'):   ('capi4linux',),
                ('suse11', '64'):   ('capi4linux-32bit',),
                ('suse', '32'):     ('libcapi20-3',),
                ('suse', '64'):     ('libcapi20-3-32bit',),
                },
          },
     'missinglibcups':
         {'packages': {
                ('debian6', '32'):  ('libcups2',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libcups2:i386',),
                ('fedora', '32'):   ('cups-libs',),
                ('fedora', '64'):   ('cups-libs.i686', 'cups-libs.x86_64'),
                ('mandriva', None): ('libcups2',),
                ('suse', '32'):     ('cups-libs',),
                ('suse', '64'):     ('cups-libs-32bit',),
                },
          },
     'missinglibdbus1':
         {'packages': {
                ('debian6', '32'):  ('libdbus-1-3',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libdbus-1-3:i386',),
                ('fedora', '32'):   ('dbus-libs',),
                ('fedora', '64'):   ('dbus-libs.i686', 'dbus-libs.x86_64'),
                ('mandriva', None): ('libdbus-1_3',),
                ('suse11', '32'):   ('dbus-1',),
                ('suse11', '64'):   ('dbus-1-32bit',),
                ('suse', '32'):     ('libdbus-1-3',),
                ('suse', '64'):     ('libdbus-1-3-32bit',),
                },
          },
     'missinglibfontconfig':
         {'packages': {
                ('debian6', '32'):  ('libfontconfig1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libfontconfig1:i386',),
                ('fedora', '32'):   ('fontconfig',),
                ('fedora', '64'):   ('fontconfig.i686', 'fontconfig.x86_64'),
                ('mandriva', None): ('libfontconfig1',),
                ('suse', '32'):     ('fontconfig',),
                ('suse', '64'):     ('fontconfig-32bit',),
                },
          },
     'missinglibglu':
         {'packages': {
                ('debian6', '32'):  ('libglu1-mesa',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libglu1-mesa:i386',),
                ('fedora', '32'):   ('mesa-libGLU',),
                ('fedora', '64'):   ('mesa-libGLU.i686', 'mesa-libGLU.x86_64'),
                ('mandriva', None): ('libmesaglu1',),
                ('suse', '32'):     ('Mesa',),
                ('suse', '64'):     ('Mesa-32bit',),
                },
          },
     'missinglibgnutls':
          {'packages': {
                ('debian6', '32'):  ('libgnutls26',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libgnutls26:i386',),
                ('fedora', '32'):   ('gnutls',),
                ('fedora', '64'):   ('gnutls.i686', 'gnutls.x86_64'),
                ('mandriva', None): ('libgnutls26',),
                ('suse11', '32'):   ('libgnutls26',),
                ('suse11', '64'):   ('libgnutls26-32bit',),
                ('suse', '32'):     ('libgnutls28',),
                ('suse', '64'):     ('libgnutls28-32bit',),
                },
          },
     'missinglibgphoto2':
         {'packages': {
                # CentOS 6 and 6 only have version 2.4
                # Debian 6 and 7 only have version 2.4
                ('fedora17', None): tuple(), # Only has version 2.4
                ('fedora', '32'):   ('libgphoto2',),
                ('fedora', '64'):   ('libgphoto2.i686', 'libgphoto2.x86_64'),
                ('mandriva', None): ('libgphoto2',),
                ('suse12', None):   tuple(), # Only has version 2.4
                ('suse', '32'):     ('libgphoto2-6',),
                ('suse', '64'):     ('libgphoto2-6-32bit',),
                },
          },
     'missinglibgsm':
         {'packages': {
                ('debian6', '32'):  ('libgsm1',),
                # Not available in Debian 6 64 bit
                ('debian7', None):  ('libgsm1:i386',),
                ('fedora', '32'):   ('gsm',),
                ('fedora', '64'):   ('gsm.i686', 'gsm.x86_64'),
                ('mandriva', None): ('libgsm1',),
                ('suse', '32'):     ('libgsm1',),
                ('suse', '64'):     ('libgsm1-32bit',),
                ('ubuntu11.10', '32'): ('libgsm1',),
                ('ubuntu11.10', '64'): ('ia32-libs',),
                },
          },
     'missinglibgstapp':
         {'packages': {
                ('debian6', '32'):  ('libgstreamer-plugins-base0.10-0',),
                ('debian7', None):  ('libgstreamer-plugins-base0.10-0:i386',),
                ('fedora', '32'):   ('gstreamer-plugins-base',),
                ('fedora', '64'):   ('gstreamer-plugins-base.i686',
                                     'gstreamer-plugins-base.x86_64'),
                ('mandriva', None): ('libgstreamer-plugins-base0.10',),
                ('suse', '32'):     ('gstreamer-0_10-plugins-base',),
                ('suse', '64'):     ('gstreamer-0_10-plugins-base-32bit',),
                },
          },
     'missinglibgstreamer':
         {'packages': {
                ('debian6', '32'):  ('libgstreamer0.10-0',),
                ('debian7', None):  ('libgstreamer0.10-0:i386',),
                ('fedora', '32'):   ('gstreamer',),
                ('fedora', '64'):   ('gstreamer.i686', 'gstreamer.x86_64'),
                ('mandriva', None): ('libgstreamer0.10_0.10',),
                ('suse', '32'):     ('libgstreamer-0_10-0',),
                ('suse', '64'):     ('libgstreamer-0_10-0-32bit',),
                },
          },
     'missinglibhal':
         {'packages': {
                ('centos7', None):  tuple(), # Not available
                ('debian6', '32'):  ('libhal1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libhal1:i386',),
                ('fedora15', '32'): ('hal-libs',), # Dropped in Fedora 16
                ('fedora15', '64'): ('hal-libs.i686','hal-libs.x86_64'),
                ('mandriva', None): ('libhal1',),
                ('suse11', '32'):   ('hal',), # Dropped in SUSE 12
                ('suse11', '64'):   ('hal-32bit',),
                ('ubuntu13', None): tuple(),
                },
          },
     'missingliblcms2':
         {'packages': {
                ('centos6', None):  tuple(), # Only has version 1.x
                # Not available in Debian 6
                ('debian7', None):  ('liblcms2-2:i386',),
                ('fedora13', None): tuple(), # Only has version 1.x
                ('fedora', '32'):   ('lcms2',),
                ('fedora', '64'):   ('lcms2.i686', 'lcms2.x86_64'),
                ('suse11.3', None): tuple(), # Not available
                ('suse', '32'):     ('liblcms2-2',),
                ('suse', '64'):     ('liblcms2-2-32bit',),
                ('ubuntu11.04', None): tuple(),
                ('ubuntu11.10', '32'): ('liblcms2-2',),
                ('ubuntu11.10', '64'): ('liblcms2-2:i386',),
                },
          },
     'missinglibldap_r':
         {'packages': {
                ('debian6', '32'):  ('libldap-2.4-2',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libldap-2.4-2:i386',),
                ('fedora', '32'):   ('openldap',),
                ('fedora', '64'):   ('openldap.i686', 'openldap.x86_64'),
                ('mandriva', None): ('libldap2.4_2',),
                ('suse', '32'):     ('libldap-2_4-2',),
                ('suse', '64'):     ('libldap-2_4-2-32bit',),
                },
          },
     'missinglibmpg123':
         {'packages': {
                ('debian6', '32'):  ('libmpg123-0',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libmpg123-0:i386',),
                ('mandriva', None): ('libmpg123_0',),
                },
          },
     'missinglibopenal':
         {'packages': {
                ('centos6', None):  tuple(), # Not available
                ('centos7', None):  tuple(), # Not available
                ('debian6', '32'):  ('libopenal1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libopenal1:i386',),
                ('fedora', '32'):   ('openal-soft',),
                ('fedora', '64'):   ('openal-soft.i686', 'openal-soft.x86_64'),
                ('mandriva', None): ('libopenal1',),
                ('suse12', '32'):   ('libopenal1-soft',),
                ('suse12', '64'):   ('libopenal1-soft-32bit',),
                ('suse', '32'):     ('libopenal1',),
                ('suse', '64'):     ('libopenal1-32bit',),
                },
          },
     'missinglibosmesa':
         {'packages': {
                # CentOS 7 only has OSMesa >= .8
                ('debian6', '32'):  ('libosmesa6',),
                ('debian7', None):  ('libosmesa6:i386',),
                ('fedora14', '32'): ('mesa-libOSMesa',),
                ('fedora14', '64'): ('mesa-libOSMesa.i686',
                                     'mesa-libOSMesa.x86_64'),
                # Fedora >= 15 only has OSMesa >= .7
                # SUSE >= 11.4 only has OSMesa >= .7
                ('ubuntu11.04', '64'): tuple(),
                ('ubuntu11.10', '32'): ('libosmesa6',),
                ('ubuntu11.10', '64'): ('libosmesa6:i386',),
                },
          },
     'missinglibsane':
         {'packages': {
                ('debian6', '32'):  ('libsane',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libsane:i386',),
                ('fedora', '32'):   ('sane-backends-libs',),
                ('fedora', '64'):   ('sane-backends-libs.i686',
                                     'sane-backends-libs.x86_64'),
                ('mandriva', None): ('libsane1',),
                ('suse', '32'):     ('sane-backends',),
                ('suse', '64'):     ('sane-backends-32bit',),
                },
          },
     'missinglibssl':
         {'packages': {
                ('debian6', '32'):  ('libssl0.9.8',),
                ('debian7', None):  ('libssl1.0.0:i386',),
                ('fedora17', '32'): ('openssl',),
                ('fedora17', '64'): ('openssl.i686', 'openssl.x86_64'),
                ('fedora', '32'):   ('openssl-libs',),
                ('fedora', '64'):   ('openssl-libs.i686',
                                     'openssl-libs.x86_64'),
                },
          },
     'missinglibtiff':
         {'packages': {
                # CentOS 6 only has version 3.9
                # CentOS 7 only has version 5
                ('debian6', '32'):  ('libtiff4',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libtiff4:i386',),
                # Fedora <= 17 only has version <= 3.9
                # Fedora >= 18 only has version >= 5
                # SUSE >= 12.2 only has version >= 5
                ('ubuntu13', None): tuple(), # Only has version >= 5
                },
          },
     'missinglibv4l1':
         {'packages': {
                ('debian6', '32'):  ('libv4l1-0',),
                # Not available on Debian 6 64 bit systems
                ('debian7', None):  ('libv4l-0:i386',),
                ('fedora', '32'):   ('libv4l1',),
                ('fedora', '64'):   ('libv4l1.i686', 'libv4l1.x86_64'),
                ('mandriva', None): ('libv4l0',),
                ('suse11.2', '32'): ('libv4l',),
                ('suse11.2', '64'): ('libv4l-32bit',),
                ('suse11', '32'):   ('libv4l1-0',),
                ('suse11', '64'):   ('libv4l1-0-32bit',),
                ('suse', '32'):     ('libv4l',),
                ('suse', '64'):     ('libv4l-32bit',),
                ('ubuntu11.10', '32'): ('libv4l-0',),
                # Not available on Ubuntu <= 11 64 bit systems
                },
          },
     'missinglibxcomposite':
         {'packages': {
                ('debian6', '32'):  ('libxcomposite1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxcomposite1:i386',),
                ('fedora', '32'):   ('libXcomposite',),
                ('fedora', '64'):   ('libXcomposite.i686',
                                     'libXcomposite.x86_64'),
                ('mandriva', None): ('libxcomposite1',),
                ('suse11', '32'):   ('xorg-x11-libs',),
                ('suse11', '64'):   ('xorg-x11-libs-32bit',),
                ('suse', '32'):     ('libXcomposite1',),
                ('suse', '64'):     ('libXcomposite1-32bit',),
                },
          },
     'missinglibxcursor':
         {'packages': {
                ('debian6', '32'):  ('libxcursor1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxcursor1:i386',),
                ('fedora', '32'):   ('libXcursor',),
                ('fedora', '64'):   ('libXcursor.i686', 'libXcursor.x86_64'),
                ('mandriva', None): ('libxcursor1',),
                ('suse11', '32'):   ('xorg-x11-libs',),
                ('suse11', '64'):   ('xorg-x11-libs-32bit',),
                ('suse', '32'):     ('libXcursor1',),
                ('suse', '64'):     ('libXcursor1-32bit',),
                },
          },
     'missinglibxi':
         {'packages': {
                ('debian6', '32'):  ('libxi6',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxi6:i386',),
                ('fedora', '32'):   ('libXi',),
                ('fedora', '64'):   ('libXi.i686', 'libXi.x86_64'),
                ('mandriva', None): ('libxi6',),
                ('suse11.2', '32'): ('xorg-x11-libs',),
                ('suse11.2', '64'): ('xorg-x11-libs-32bit',),
                ('suse', '32'):     ('libXi6',),
                ('suse', '64'):     ('libXi6-32bit',),
                },
          },
     'missinglibxinerama':
         {'packages': {
                ('debian6', '32'):  ('libxinerama1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxinerama1:i386',),
                ('fedora', '32'):   ('libXinerama',),
                ('fedora', '64'):   ('libXinerama.i686', 'libXinerama.x86_64'),
                ('mandriva', None): ('libxinerama1',),
                ('suse11', '32'):   ('xorg-x11-libs',),
                ('suse11', '64'):   ('xorg-x11-libs-32bit',),
                ('suse', '32'):     ('libXinerama1',),
                ('suse', '64'):     ('libXinerama1-32bit',),
                },
          },
     'missinglibxml2':
         {'packages': {
                ('debian6', '32'):  ('libxml2',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxml2:i386',),
                ('fedora', '32'):   ('libxml2',),
                ('fedora', '64'):   ('libxml2.i686', 'libxml2.x86_64'),
                ('mandriva', None): ('libxml2_2',),
                ('suse11', '32'):   ('libxml2',),
                ('suse11', '64'):   ('libxml2-32bit',),
                ('suse', '32'):     ('libxslt1',),
                ('suse', '64'):     ('libxslt1-32bit',),
                },
          },
     'missinglibxrandr':
         {'packages': {
                ('debian6', '32'):  ('libxrandr2',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxrandr2:i386',),
                ('fedora', '32'):   ('libXrandr',),
                ('fedora', '64'):   ('libXrandr.i686', 'libXrandr.x86_64'),
                ('mandriva', None): ('libxrandr2',),
                ('suse11', '32'):   ('xorg-x11-libs',),
                ('suse11', '64'):   ('xorg-x11-libs-32bit',),
                ('suse', '32'):     ('libXrandr2',),
                ('suse', '64'):     ('libXrandr2-32bit',),
                },
          },
     'missinglibxrender':
         {'packages': {
                ('debian6', '32'):  ('libxrender1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxrender1:i386',),
                ('fedora', '32'):   ('libXrender',),
                ('fedora', '64'):   ('libXrender.i686', 'libXrender.x86_64'),
                ('mandriva', None): ('libxrender1',),
                ('suse11', '32'):   ('xorg-x11-libXrender',),
                ('suse11', '64'):   ('xorg-x11-libXrender-32bit',),
                ('suse', '32'):     ('libXrender1',),
                ('suse', '64'):     ('libXrender1-32bit',),
                },
          },
     'missinglibxslt':
         {'packages': {
                ('debian6', '32'):  ('libxslt1.1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxslt1.1:i386',),
                ('fedora', '32'):   ('libxslt',),
                ('fedora', '64'):   ('libxslt.i686', 'libxslt.x86_64'),
                ('mandriva', None): ('libxslt1',),
                ('suse11', '32'):   ('libxslt',),
                ('suse11', '64'):   ('libxslt-32bit',),
                ('suse', '32'):     ('libxslt1',),
                ('suse', '64'):     ('libxslt1-32bit',),
                },
          },
     'missinglibxxf86vm':
         {'packages': {
                ('debian6', '32'):  ('libxxf86vm1',),
                ('debian6', '64'):  ('ia32-libs',),
                ('debian7', None):  ('libxxf86vm1:i386',),
                ('fedora', '32'):   ('libXxf86vm',),
                ('fedora', '64'):   ('libXxf86vm.i686', 'libXxf86vm.x86_64'),
                ('mandriva', None): ('libxxf86vm1',),
                ('suse11', '32'):   ('xorg-x11-libs',),
                ('suse11', '64'):   ('xorg-x11-libs-32bit',),
                ('suse', '32'):     ('libXxf86vm1',),
                ('suse', '64'):     ('libXxf86vm1-32bit',),
                },
          },
     'missingliblibz':
         {'packages': {
                ('debian6', '32'):  ('zlib1g',),
                ('debian6', '64'):  ('lib32z1',),
                ('debian7', None):  ('zlib1g:i386',),
                ('fedora', '32'):   ('zlib',),
                ('fedora', '64'):   ('zlib.i686', 'zlib.x86_64'),
                ('mandriva', None): ('zlib1',),
                ('suse', '32'):     ('libz1',),
                ('suse', '64'):     ('libz1-32bit',),
                },
          },
     'openglnotextures3tc':
         {'packages': {
                ('debian7', None):  ('libtxc-dxtn-s2tc0:i386',),
                },
          },
     }


DISTROS = \
    {'centos6':   {'name': 'CentOS <= 6',
                   'globs': {'etc/redhat-release': ('CentOS release [1-6]\\.',)},
                   'priority': 10,
                   'packagecmd': 'yum install',
                   'fallbacks': ('fedora14',),
                   },
     'centos7':   {'name': 'CentOS 7',
                   'globs': {'etc/redhat-release': ('CentOS Linux release 7\\.',)},
                   'priority': 10,
                   'packagecmd': 'yum install',
                   'fallbacks': ('fedora',),
                   },
     'debian6':   {'name': 'Debian <= 6',
                   'globs': {'etc/debian_version': ('[1-6]\\.',)},
                   'priority': 10,
                   'updatecmd': 'apt-get update',
                   'packagecmd': 'apt-get install',
                   },
     'debian7':   {'name': 'Debian',
                   'globs': {'etc/os-release': ('ID *= *"?debian',),
                             'etc/debian_version': tuple()},
                   'updatecmd': 'apt-get update',
                   'packagecmd': 'apt-get install',
                   },
     'fedora13':  {'name': 'Fedora <= 13',
                   'globs': {'etc/fedora-release': ('Fedora release ([0-9]|1[0-3]) ',)},
                   'priority': 10,
                   'packagecmd': 'yum install',
                   'fallbacks': ('fedora14',),
                   },
     'fedora14':  {'name': 'Fedora <= 14',
                   'globs': {'etc/fedora-release': ('Fedora release 14 ',)},
                   'priority': 10,
                   'packagecmd': 'yum install',
                   'fallbacks': ('fedora15',),
                   },
     'fedora15':  {'name': 'Fedora <= 15',
                   'globs': {'etc/fedora-release': ('Fedora release 15 ',)},
                   'priority': 10,
                   'packagecmd': 'yum install',
                   'fallbacks': ('fedora17',),
                   },
     'fedora17':  {'name': 'Fedora <= 17',
                   'globs': {'etc/fedora-release': ('Fedora release 1[67]',)},
                   'priority': 10,
                   'packagecmd': 'yum install',
                   'fallbacks': ('fedora',),
                   },
     'fedora':    {'name': 'Fedora / Red Hat',
                   'globs': {'etc/fedora-release': tuple(),
                             'etc/redhat-release': tuple()},
                   'packagecmd': 'yum install',
                   },
     'mandriva':  {'name': 'Mandriva',
                   'globs': {'etc/mandriva-release': tuple()},
                   'priority': 50, # Typically fedora will also match
                   'packagecmd': '/usr/sbin/urpmi', # Usually not in $PATH
                   },
     'suse11.2':   {'name': 'openSUSE <= 11.2',
                   'globs': {'etc/SuSE-release': ('VERSION *= *([1-9]\\.|10\\.|11\\.[012])',)},
                   'priority': 15,
                   'packagecmd': 'zypper install',
                   'fallbacks': ('suse11.3',),
                   },
     'suse11.3':   {'name': 'openSUSE <= 11.3',
                   'globs': {'etc/SuSE-release': ('VERSION *= *11\\.3',)},
                   'priority': 15,
                   'packagecmd': 'zypper install',
                   'fallbacks': ('suse11',),
                   },
     'suse11':     {'name': 'openSUSE <= 11',
                   'globs': {'etc/SuSE-release': ('VERSION *= *11\\.',)},
                   'priority': 10,
                   'packagecmd': 'zypper install',
                   'fallbacks': ('suse12',),
                   },
     'suse12':     {'name': 'openSUSE <= 12',
                   'globs': {'etc/SuSE-release': ('VERSION *= *12\\.',)},
                   'priority': 10,
                   'packagecmd': 'zypper install',
                   'fallbacks': ('suse',),
                   },
     'suse':      {'name': 'openSUSE',
                   'globs': {'etc/os-release': ('ID *= *"?opensuse',),
                             'etc/SuSE-release': tuple()},
                   'packagecmd': 'zypper install',
                   },
     'ubuntu11.04': {'name': 'Ubuntu <= 11.04',
                   'globs': {'etc/lsb-release':
                                 ('DISTRIB_ID *= *Ubuntu',
                                  'DISTRIB_RELEASE *= *([0-9]\\.|10\\.|11\\.04)')},
                   'priority': 20, # Typically debian[67] will also match
                   'updatecmd': 'apt-get update',
                   'packagecmd': 'apt-get install',
                   'fallbacks': ('ubuntu11.10',),
                   },
     'ubuntu11.10': {'name': 'Ubuntu <= 11.10',
                   'globs': {'etc/lsb-release':
                                 ('DISTRIB_ID *= *Ubuntu',
                                  'DISTRIB_RELEASE *= *11\\.10')},
                   'priority': 20, # Typically debian[67] will also match
                   'updatecmd': 'apt-get update',
                   'packagecmd': 'apt-get install',
                   'fallbacks': ('debian6',),
                   },
     'ubuntu13':  {'name': 'Ubuntu <= 13.10',
                   'globs': {'etc/lsb-release':
                                 ('DISTRIB_ID *= *Ubuntu',
                                  'DISTRIB_RELEASE *= *1[23]\\.')},
                   'priority': 20, # Typically debian[67] will also match
                   'updatecmd': 'apt-get update',
                   'packagecmd': 'apt-get install',
                   'fallbacks': ('debian7',),
                   },
     'ubuntu':    {'name': 'Ubuntu',
                   'globs': {'etc/lsb-release': ('DISTRIB_ID *= *"?Ubuntu',),
                             'etc/os-release': ('ID *= *"?ubuntu',)},
                   'priority': 15, # Typically debian[67] will also match
                   'updatecmd': 'apt-get update',
                   'packagecmd': 'apt-get install',
                   'fallbacks': ('ubuntu13',),
                   },
     }


def display_error_zenity(text):
    if isinstance(text, cxutils.unicode_type):
        text = text.encode('utf8')
    return cxutils.run(['zenity', '--error', '--text', text])

def display_error_kdialog(text):
    if isinstance(text, cxutils.unicode_type):
        text = text.encode('utf8')
    return cxutils.run(['kdialog', '--error', text])

def display_error_xmessage(text):
    if isinstance(text, cxutils.unicode_type):
        text = text.encode('utf8') # FIXME
    return cxutils.run(['xmessage', text])

def display_error(text):
    retcode, _out, _err = display_error_zenity(text)
    if retcode < 0:
        retcode, _out, _err = display_error_kdialog(text)
    if retcode < 0:
        retcode, _out, _err = display_error_xmessage(text)
    if retcode < 0:
        sys.stderr.write(text)


_BITNESS = None

def detect_bitness():
    # pylint: disable-msg=W0603
    global _BITNESS
    if _BITNESS is None:
        _BITNESS = '64'
        # We don't detect bitness through uname to correctly deal with chroots.
        # Ignore _retcode: it will indicate an error if either files is missing.
        _retcode, out, _err = cxutils.run(('file', '/bin/ls', '/usr/bin/ls'), stdout=cxutils.GRAB, stderr=cxutils.NULL)
        if out.find('ELF 32-bit') >= 0:
            _BITNESS = '32'
    return _BITNESS

_DISTROID = None

def detect_distribution():
    # pylint: disable-msg=W0603
    global _DISTROID
    if _DISTROID is None:
        glob_tree = globtree.FileContentGlobTree()

        for (distroid, info) in DISTROS.iteritems():
            for (file_glob, patterns) in info['globs'].iteritems():
                try:
                    glob_tree.add_content_glob(file_glob, patterns, distroid)
                except re.error:
                    cxlog.warn("the %s fileglob of %s contains an invalid regular expression" % (file_glob, distroid))


        best_priority = -1
        for _filename, distroid in glob_tree.matches('/'):
            if _DISTROID is None or \
                    best_priority < DISTROS[distroid].get('priority', 0):
                _DISTROID = distroid
                best_priority = DISTROS[distroid].get('priority', 0)
                cxlog.log("found a match for distro %s (%s)" % (_DISTROID, best_priority))
    return _DISTROID


_ERRORS = {}

def has_fix(errid):
    global FIXES
    if errid not in FIXES:
        return False

    distroid = detect_distribution()
    if distroid is None:
        return False

    bitness = detect_bitness()
    if get_packages(errid, distroid, bitness):
        return True

    return False


def add_error(errid, description=None):
    if description is None:
        global FIXES
        if errid in FIXES:
            description = FIXES[errid]['']
        else:
            description = _('Unknown error %s' % errid)
    global _ERRORS
    _ERRORS[errid] = description

def remove_error(errid):
    global _ERRORS
    del _ERRORS[errid]

def has_error(errid):
    global _ERRORS
    return errid in _ERRORS

def has_errors():
    global _ERRORS
    if _ERRORS:
        return True
    return False

def clear_errors():
    # pylint: disable-msg=W0603
    global _ERRORS
    _ERRORS = {}

def get_packages(errid, distroid, bitness):
    global FIXES
    if errid not in FIXES:
        return None

    global DISTROS
    distroids = []
    newids = [distroid]
    while newids:
        distroid = newids.pop(0)
        distroids.append(distroid)
        for newid in DISTROS[distroid].get('fallbacks', tuple()):
            if newid not in distroids:
                newids.append(newid)

    # Check if there are any other errors that this one may have masked and
    # which should be fixed at the same time
    errids = [errid]
    for fixid, fix in FIXES.iteritems():
        if errid in fix.get('tiedto', []):
            errids.append(fixid)

    # Return the packages for all the relevant errors
    packages = []
    for errid in errids:
        for distroid in distroids:
            fix_packages = FIXES[errid]['packages']
            if (distroid, bitness) in fix_packages:
                packages.extend(fix_packages[(distroid, bitness)])
                break
            if (distroid, None) in fix_packages:
                packages.extend(fix_packages[(distroid, None)])
                break
    return packages

def fix_errors():
    if not has_errors():
        return

    distroid = detect_distribution()
    if distroid is None:
        return
    bitness = detect_bitness()

    fixable_errors = set()
    packages = set()
    global _ERRORS
    for errid in _ERRORS:
        fix_packages = get_packages(errid, distroid, bitness)
        if fix_packages:
            fixable_errors.add(errid)
            packages.update(fix_packages)

    if fixable_errors:
        cmd = DISTROS[distroid]['packagecmd'] + ' ' + cxutils.argvtocmdline(sorted(packages))
        if 'updatecmd' in DISTROS[distroid]:
            # Ignore update errors, hoping the package list was not too outdated
            cmd = DISTROS[distroid]['updatecmd'] + "; " + cmd

        # If we need to install a Debian multiarch package, make sure we have
        # the required architecture
        for package in packages:
            if ':' in package:
                retcode, out, _err = cxutils.run(('dpkg', '--print-architecture'), stdout=cxutils.GRAB, stderr=cxutils.NULL)
                if retcode == 0 and out != 'i386\n':
                    retcode, out, _err = cxutils.run(('dpkg', '--print-foreign-architectures'), stdout=cxutils.GRAB, stderr=cxutils.NULL)
                    if retcode == 0 and out.find('i386\n') < 0:
                        cmd = 'dpkg --add-architecture i386 && (' + cmd + ')'
                break

        # The packaging tools are likely to ask for confirmation on the console
        cxsu_cmd = [os.path.join(cxutils.CX_ROOT, "bin", "cxsu"), '--console', '--ignore-home', 'sh', '-c', cmd]
        retcode, _out, _err = cxutils.run(cxsu_cmd)
        if retcode == 0:
            for errid in fixable_errors:
                remove_error(errid)

def report_errors():
    if not has_errors():
        return

    message = _("Some errors may prevent %s from working correctly:\n") % distversion.PRODUCT_NAME
    global _ERRORS
    for description in sorted(_ERRORS.values()):
        message += "* " + description + "\n"

    packages = {}
    for errid in _ERRORS.keys():
        if errid not in FIXES:
            continue
        fix_packages = FIXES[errid]['packages']
        for key in packages.keys():
            if key not in packages:
                packages[key] = set()
            packages[key].update(fix_packages[key])

    if packages:
        message += _("\nYou may be able to fix them by running one of the following commands as root:\n")
        lines = []
        for (distroid, bitness) in packages.keys():
            name = DISTROS[distroid]['name']
            if bitness is not None:
                name = _("%(name)s (%(bitness)s bits)") % {'name': name, 'bitness': bitness}
            lines.append("%-23s\t%s %s\n" % (name, DISTROS[distroid]['cmd'],
                                             ' '.join(sorted(packages[(distroid, bitness)]))))
        message += ''.join(sorted(lines))
    display_error(message)
